# -*- coding: UTF-8 -*-
###########################################################################
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# lib_backend.py
#
# classes de base pour la gestion des serveurs sous zephir
#
###########################################################################
"""fonctions utiles pour l'accès à la base de données (via adbapi)
"""
from zephir.backend import config
from zephir.backend.config import log
try:
    from migration_perso import migration_perso, destinations_perso
except:
    migration_perso = {}
    destinations_perso = []
from zephir.utils.creolewrap import ZephirDict
from zephir.monitor.agentmanager.util import md5files
from twisted.internet.utils import getProcessOutputAndValue
import psycopg2 as PgSQL
import os, sys, time, string, traceback, apt_pkg, re, shutil
import base64, md5
from ConfigParser import ConfigParser
from glob import glob
from cStringIO import StringIO
import xmlrpclib

class ResourceAuthError(Exception):
    pass

class CxPool(dict):
    """dictionnaire des connexions postgresql en cours"""

    def create(self):
        cx = PgSQL.connect(database=config.DB_NAME,user=config.DB_USER,password=config.DB_PASSWD)
        cu = cx.cursor()
        self[cu] = cx
        return cu

    def commit(self,cu):
        cu.close()
        self[cu].commit()
        self[cu].close()
        del(self[cu])

    def rollback(self,cu):
        cu.close()
        self[cu].rollback()
        self[cu].close()
        del(self[cu])

    def close(self,cu):
        cu.close()
        self[cu].close()
        del(self[cu])

class AptChecker:
    """gestionnaire de vérification d'état des paquets d'un serveur
    """
    def __init__(self):
        self.pkg_file = {}
        self.packages = {}
        self.timestamp = {}
        for zephir_version, data in config.DISTRIBS.items():
            if zephir_version != 1:
                codename, version, maintained = data
                # lecture des fichiers générés pour les différents serveurs de mise à jour
                for serv_maj in config.SERVEURS_MAJ_CLIENTS:
                    if zephir_version not in self.pkg_file:
                        self.pkg_file[zephir_version] = {}
                        self.packages[zephir_version] = {}
                        self.timestamp[zephir_version] = {}
                    self.pkg_file[zephir_version][serv_maj] = os.path.join(os.path.abspath(config.PATH_ZEPHIR),'packages_%s_%s.ini' % (codename, serv_maj))
                    self.packages[zephir_version][serv_maj] = ConfigParser()
                    if not os.path.isfile(self.pkg_file[zephir_version][serv_maj]):
                        f = file(self.pkg_file[zephir_version][serv_maj],'w')
                        f.close()
                        # la liste des paquets n'a jamais été vérifiée, on lance la procédure de recherche
                        log.msg("** récupération de la liste des paquets disponibles en mise à jour", serv_maj)
                        os.system('/usr/share/eole/check_maj_eole.py')
                    self.timestamp[zephir_version][serv_maj] = os.path.getmtime(self.pkg_file[zephir_version][serv_maj])
                    self.packages[zephir_version][serv_maj].read(self.pkg_file[zephir_version][serv_maj])
        # création d'une regexp compilée pour les paquets à ignorer
        if len(config.held_packages) > 0:
            self.held_pattern = re.compile("|".join(config.held_packages))
        else:
            self.held_pattern = None

    def check_min_version(self, current_version, min_version):
        apt_pkg.InitSystem()
        if apt_pkg.VersionCompare(current_version, min_version) >= 0:
            return True
        return False

    def check_packages(self, maj_infos, pkg_file, debnames, version, serv_maj, type_maj = 'complete', show_installed = False):
        """vérifie l'état de mise à jour des paquets
        type_maj : type de mise à jour (minimale/complete)
        show_installed : si True, renvoie la liste de tous les paquets installés
                         par défaut, seulement les paquets non à jour
        """
        # si serveur de maj inconnu, on utilise le premier de la liste
        if serv_maj not in self.packages[version]:
            config.log.msg("Serveur de mise à jour", serv_maj, "inconnu : utilisation de", config.SERVEURS_MAJ_CLIENTS[0])
            serv_maj = config.SERVEURS_MAJ_CLIENTS[0]
        depot_changed = True
        client_changed = True
        if os.path.isfile(pkg_file):
            md5_packs = md5.md5(file(pkg_file).read()).hexdigest()
        else:
            # pas de fichier remonté par le serveur !
            if maj_infos is not None:
                return maj_infos[0], maj_infos
            else:
                return [], None

        depot_timestamp = os.path.getmtime(self.pkg_file[version][serv_maj])
        if maj_infos:
            # on regarde si les données ont changé depuis la dernière vérification de ce serveur
            if maj_infos[1] == type_maj and maj_infos[2] == md5_packs and maj_infos[3] == serv_maj:
                client_changed = False
            if maj_infos[4] == depot_timestamp:
                depot_changed = False

        if debnames == [] and not show_installed:
            # vérification globale de l'état de mise à jour du serveur
            # si on a déjà des infos et que rien n'a changé, on ne fait pas de traitements
            if not depot_changed and not client_changed:
                return maj_infos[0], maj_infos

        # creation de la liste des paquets installés (/à vérifier si debnames non vide)
        pkg_liste = []
        if os.path.isfile(pkg_file):
            for line in file(pkg_file):
                pak_info = line.split()
                if debnames != []:
                    # travail sur des paquets particuliers
                    if pak_info[0] in debnames:
                        pkg_liste.append(pak_info)
                        if len(pkg_liste) == len(debnames):
                            # on a trouvé tous les paquets demandés : on sort
                            break
                else:
                    pkg_liste.append(line.split())
        # md5 des paquets installés
        check_needed = False
        if maj_infos is None or client_changed == True or debnames != [] or show_installed == True:
            # Des informations ne sont pas à jour ou on a demandé les infos sur des paquets particuliers
            check_needed = True
        if depot_changed:
            # la liste des paquets dispos a changé, on la recharge
            self.packages[version][serv_maj].read(self.pkg_file[version][serv_maj])
            self.timestamp[version][serv_maj] = depot_timestamp
            check_needed = True
        liste_pack = []
        if check_needed:
            outdated = []
            apt_pkg.InitSystem()
            for pkg_name, pkg_version in pkg_liste:
                if (self.held_pattern is None) or (not self.held_pattern.match(pkg_name)):
                    try:
                        # les paquets n'ayant jamais eu de mise à jour ne sont pas présents
                        if self.packages[version][serv_maj].has_section(pkg_name):
                            try:
                                maj_version = self.packages[version][serv_maj].get(pkg_name, type_maj)
                            except:
                                if type_maj != 'minimum':
                                    log.msg('Informations de version du paquet %s non trouvées pour %s' % (pkg_name, serv_maj))
                                    traceback.print_exc()
                            if apt_pkg.VersionCompare(maj_version, pkg_version) > 0:
                                # on renvoie, le paquet, la version installée, et la version sur le serveur de maj
                                liste_pack.append((pkg_name, pkg_version, maj_version))
                                outdated.append((pkg_name, pkg_version, maj_version))
                                continue
                            elif show_installed == True:
                                liste_pack.append((pkg_name, pkg_version, maj_version))
                                continue
                    except:
                        traceback.print_exc()
                    # listing des paquets installés, on renvoie le paquet
                    # même si il n'est pas dans les repository de maj
                    if show_installed == True:
                        liste_pack.append((pkg_name, pkg_version, pkg_version))
            maj_infos = (outdated, type_maj, md5_packs, serv_maj, depot_timestamp)
        else:
            # pas de changement, on reprend la liste précédente
            liste_pack = maj_infos[0]
        return liste_pack, maj_infos

cx_pool = CxPool()
maj_checker = AptChecker()

class Serveur:
    """classe utilitaire pour récupérer diverses données sur un serveur"""

    def __init__(self, pool, id_s, cu=None, data=None):
        """initialise l'objet serveur et récupère les infos
        """
        self.version = 'creole2'
        # récupération des serveurs
        self.id_s = int(id_s)
        # date de dernière modification du serveur
        self.modified = time.time()
        # date de création de l'objet
        self.created = time.time()
        # serveur inactif ? (pas d'alerte)
        self.no_alert = False
        # lien sur le pool de serveurs
        self.pool = pool
        # liste des paquets non à jour
        self.maj_infos = None
        cursor = cu
        if cu == None:
            cursor = cx_pool.create()
        try:
            if data == None:
                # le rne ne peut pas être modifié : on le stocke une fois pour toutes
                cursor.execute("select rne from serveurs where id=%s", (int(self.id_s),))
                data = cursor.fetchone()
                self.rne = data[0]
                # appel de la fonction update_data pour les données modifiables
                self.update_data(cursor)
            else:
                self.rne = data[0]
            self.confdir = os.path.join(config.PATH_ZEPHIR,'conf',self.rne, str(self.id_s))
            if cu == None:
                cx_pool.close(cursor)
        except:
            traceback.print_exc()
            if cu == None:
                cx_pool.close(cursor)

    def update_data(self, cu=None, data=None):
        """recharge les données des serveurs
        """
        # récupération de l'établissement et de l'état
        cursor = cu
        if cu == None:
            cursor = cx_pool.create()
        try:
            if data == None:
                cursor.execute("select libelle, module_actuel, variante, timeout, etat, md5s, maj, no_alert from serveurs where id=%s", (int(self.id_s),))
                data = cursor.fetchone()
            # données du serveur
            self.libelle = data[0]
            self.id_mod = int(data[1])
            self.id_var = int(data[2])
            try:
                self.md5s = int(data[5])
            except:
                self.md5s = -1
            try:
                self.maj = int(data[6])
            except:
                self.maj = -1
            try:
                self.timeout = int(data[3])
            except:
                self.timeout = 0
            if data[4] == None:
                self.status = -1
            else:
                self.status = int(data[4])
            try:
                if str(data[7]) == '1':
                    self.no_alert = True
                else:
                    self.no_alert = False
            except:
                pass
            # libelle etablissement
            cursor.execute("select libelle from etablissements where rne=%s", (self.rne,))
            data = cursor.fetchone()
            self.etab = data[0]
            # libelle module
            cursor.execute("select libelle,version from modules where id=%s", (int(self.id_mod),))
            data = cursor.fetchone()
            self.module = data[0]
            try:
                self.module_version = int(data[1])
            except:
                self.module_version = 4
            # détection des modules utilisant creole1
            OLD_MODULES =  config.liste_modules[1]
            if self.module in OLD_MODULES:
                self.version = 'creole1'
            else:
                self.version = 'creole2'
            # libelle variante
            cursor.execute("select libelle from variantes where id=%s", (int(self.id_var),))
            data = cursor.fetchone()
            self.variante = data[0]
            self.modified = time.time()
            if cu == None:
                cx_pool.close(cursor)
        except:
            traceback.print_exc()
            if cu == None:
                cx_pool.close(cursor)

    def __repr__(self):
        """affichage du serveur
        """
        return """%s => %s (%s)""" % (self.id_s, self.module, self.rne)

    def __str__(self):
        """affichage du serveur
        """
        return """%s => %s (%s) - %s (%s)""" % (self.id_s, self.libelle, self.module, self.etab, self.rne)

    #####################
    ## LECTURE DE DONNEES

    def get_libelle(self):
        return self.libelle

    def get_rne(self):
        return self.rne

    def get_etab(self):
        return self.etab

    def get_module(self):
        return self.module

    def get_variante(self):
        return self.variante

    def get_timeout(self):
        return self.timeout

    def get_confdir(self):
        return self.confdir

    def get_config(self, mode="modif_config", encode=False):
        """retourne la configuration eole du serveur (zephir.eol)
        """
        # récupération du dictionnaire et des dicos locaux
        if self.version == 'creole1':
            # on recherche la liste des dictionnaires à charger
            creole_files = [self.confdir+os.sep+'dictionnaire']
            # dictionnaire de la variante
            creole_files.extend(glob(self.confdir+os.sep+'dicos/variante/*.eol'))
            # dictionnaire locaux
            creole_files.extend(glob(self.confdir+os.sep+'dicos/*.eol'))
            dicos = []
            for dic in creole_files:
                try:
                    # lecture du contenu des dictionnaires
                    fic = open(dic,'r')
                    lines = fic.readlines()
                    fic.close()
                    if encode == True:
                        data = [ unicode(line, 'ISO-8859-1').encode('UTF-8') for line in lines ]
                    else:
                        data = lines
                    # stockage du contenu du dictionnaire
                    dicos.append(data)
                except OSError:
                    pass

        if self.version == 'creole2':
            # chargement de configuration pour un serveur Eole2
            dicos = []
            for rep in ['module','variante', 'local']:
                dicos.append(os.path.join(self.confdir,'dicos',rep))

        dico = ZephirDict(dicos, self.confdir, mode, self.version)
        return dico

    def save_config(self, dico_zeph, mode='config', encode=False, force=False):
        """sauvegarde la configuration eole du serveur (zephir.eol)
        """
        try:
            if mode in ['config','modif_config']:
                fic_zephir = os.path.join(self.confdir,'zephir.eol')
                key = 'config_ok'
            elif mode in ['dico','modif_dico']:
                # si gen_dico, sauvegarde de dico.eol dans la variante
                fic_zephir = os.path.abspath(config.PATH_MODULES)+os.sep+str(self.id_mod)+os.sep+'variantes'+os.sep+str(self.id_var)+os.sep+'dico.eol'
                key = 'dico_ok'
            elif mode in ['migration','modif_migration']:
                fic_zephir = os.path.join(self.confdir,'migration.eol')
                key = 'migration_ok'
            dico_zeph.save(fic_zephir,encode=encode,force=force)
            if key != '':
                self.maj_params({key:1})
            self.check_md5conf()
        except:
            traceback.print_exc()
            raise IOError("serveur %s : erreur de sauvegarde de %s" % (self.id_s, os.path.basename(fic_zephir)))
        return ""

    def get_status(self):
        """renvoie le dernier état enregistré du serveur
        """
        try:
            assert self.status in range (5)
        except:
            return -1
        else:
            return self.status

    def set_status(self, status):
        """Met à jour l'état du serveur
        """
        self.status = int(status)
        cu = cx_pool.create()
        try:
            cu.execute("update serveurs set etat=%s where id=%s", (int(self.status), int(self.id_s)))
            cx_pool.commit(cu)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)
        return self.status

    def update_ip_pub(self, ip_publique):
        """Conserve la dernière adresse de connexion du serveur
        """
        cu = cx_pool.create()
        try:
            cu.execute("update serveurs set ip_publique=%s where id=%s", (ip_publique, int(self.id_s)))
            cx_pool.commit(cu)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)
            return False
        return True

    def get_params(self, params=None):
        """renvoie le champ paramètres du serveur
        """
        try:
            if params is None:
                cu = cx_pool.create()
                try:
                    cu.execute("select params from serveurs where id=%s", (int(self.id_s),))
                    data = cu.fetchone()
                    cx_pool.close(cu)
                except:
                    traceback.print_exc()
                    cx_pool.close(cu)
                params = {}
                try:
                    params = eval(data[0])
                    # RQE : si ajout de nouveaux champs dans params,
                    # remplacer le nom du champ ci-dessous par le dernier ajouté
                    assert type(params) == dict and params.has_key('perso_ok')
                except:
                    # met à jour le champ param si il n'est pas encore renseigné
                    params = self.update_params(params)
                    # sauvegarde du champs params
                    self.save_params(params)
            else:
                if type(params) == str:
                    params = eval(params)

            # nombre de commandes et transferts uucp en attente
            id_uucp = self.rne+'-'+str(self.id_s)
            params['uucp_transfert']=0
            params['uucp_cmd']=0
            if os.path.isdir('/var/spool/uucp/%s/D.' % id_uucp):
                params['uucp_transfert']=len(os.listdir('/var/spool/uucp/%s/D.' % id_uucp))
            if os.path.isdir('/var/spool/uucp/%s/D.X' % id_uucp):
                params['uucp_cmd']=len(os.listdir('/var/spool/uucp/%s/D.X' % id_uucp))
            # si la configuration n'était pas présénte, on vérifie si c'est toujours le cas
            if params['config_ok'] == 0:
                if os.path.isfile(self.confdir+os.sep+'zephir.eol'):
                    params['config_ok'] = 1
            if params['dico_ok'] == 0:
                if os.path.isfile(self.confdir+os.sep+'dico.eol'):
                    params['dico_ok'] = 1
            if os.path.isfile(self.confdir+os.sep+'new_key.pub'):
                if os.path.isfile(os.path.join(self.confdir,'new_addr')):
                    params['new_key'] = (2,file(os.path.join(self.confdir,'new_addr')).read().strip())
                elif os.path.isfile(os.path.join(self.confdir,'new_addr_ok')):
                    params['new_key'] = (3,file(os.path.join(self.confdir,'new_addr_ok')).read().strip())
                else:
                    params['new_key'] = (1, '')
            else:
                params['new_key'] = (0, '')
            if self.version != 'creole1':
                params['migration_ok'] = 1
            elif os.path.isfile(self.confdir+os.sep+'migration.eol'):
                params['migration_ok'] = 0
            else:
                params['migration_ok'] = -1
            return params
        except:
            traceback.print_exc()
            return {}

    def edit_serveur(self, dico_modifs):
        # on vérifie que l'identifiant n'est pas modifié
        if 'id' in dico_modifs.keys():
            return 0, config.u("""L'identifiant ne peut pas être modifié""")
        # idem pour le rne et le module
        if 'rne' in dico_modifs.keys():
            return 0, config.u("""Le rne ne peut pas être modifié""")
        if 'module_initial' in dico_modifs.keys():
            return 0, config.u("""Le module initial ne peut pas être modifié""")
        if 'module_actuel' in dico_modifs.keys() and int(dico_modifs['module_actuel']) != self.id_mod:
            # modification amon <-> amonecole permis
            mod_name = self.module[:self.module.rindex('-')]
            allowed_edits = config.allowed_mod_edits.get(self.module_version, {})
            if mod_name in allowed_edits:
                # recherche de la variante par défaut et du libellé pour le modules de destination
                query = """select modules.libelle, variantes.id, modules.version from modules, variantes where modules.id=%s and modules.id=variantes.module and variantes.libelle='standard'"""
                cu = cx_pool.create()
                try:
                    cu.execute(query, (int(dico_modifs['module_actuel']),))
                    data = cu.fetchone()
                    cx_pool.close(cu)
                except:
                    traceback.print_exc()
                    cx_pool.close(cu)
                    return 0, config.u("""Le module n'a pas été retrouvé""")
                if (not data[0][:data[0].rindex('-')] in allowed_edits[mod_name]) or (int(data[2]) != int(self.module_version)):
                    return 0, config.u("""Le module demandé n'est pas compatible avec ce serveur""")
                # cas spécial : si on modifie le module et pas la variante, mettre variante standard par défaut
                if not 'variante' in dico_modifs.keys():
                    dico_modifs['variante'] = data[1]
            else:
                return 0, config.u("""Le module ne peut pas être modifié""")
        # construction de la requête SQL de modification
        if dico_modifs == {}:
            return 1, config.u("""Aucune modification demandée""")
        else:
            params = []
            requete=["update serveurs set "]
            for cle in dico_modifs.keys():
                requete.append(str(cle))
                if cle in ('module_initial', 'module_actuel', 'variante', 'timeout', 'etat', 'maj', 'md5s', 'no_alert'):
                    requete.append("=%s, ")
                    params.append(int(dico_modifs[cle]))
                else:
                    requete.append("=E%s, ")
                    params.append(str(dico_modifs[cle]))
            query = "".join(requete)[:-2]
            query += """ where id=%s"""
            params.append(int(self.id_s))
            # modification dans la base et appel à la fonction de mise à jour dans l'arborescence
        try:
            cu = cx_pool.create()
            cu.execute(query, params)
            cx_pool.commit(cu)
        except:
            return 0, config.u("""Erreur lors de la modification dans la base""")

        serveur_dir_ori = self.confdir
        if 'variante' in dico_modifs.keys():
            if int(dico_modifs['variante']) != int(self.id_var):
                # on logue le changement de variante
                try:
                    params = (int(self.id_s), str(time.ctime()), int(self.id_var), int(dico_modifs['variante']))
                    cu = cx_pool.create()
                    cu.execute("""insert into log_serveur (id_serveur,date,type,message,etat) values (%s,%s,'ZEPHIR','Modification de la variante (%s -> %s)',0)""", params)
                    cx_pool.commit(cu)
                except:
                    # on ne bloque pas si le log échoue
                    pass

            if 'module_actuel' in dico_modifs.keys() and int(dico_modifs['module_actuel']) != self.id_mod:
                code, msg = self._cree_arbo_serveur(new_module=int(dico_modifs['module_actuel']),new_variante=int(dico_modifs['variante']))
            else:
                code, msg = self._cree_arbo_serveur(new_variante=int(dico_modifs['variante']))
            if code == 0:
                return 0, config.u(msg)

            # mise à jour des statistiques des variantes
            self.pool.stats['serv_variantes'][str(self.id_var)] = self.pool.stats['serv_variantes'][str(self.id_var)] - 1
            self.pool.stats['serv_variantes'][str(dico_modifs['variante'])] = self.pool.stats['serv_variantes'].get(str(dico_modifs['variante']), 0) + 1
        if 'module_actuel' in dico_modifs.keys() and int(dico_modifs['module_actuel']) != self.id_mod:
            # mise à jour des statistiques des modules
            self.pool.stats['serv_modules'][str(self.id_mod)] = self.pool.stats['serv_modules'][str(self.id_mod)] - 1
            self.pool.stats['serv_modules'][str(dico_modifs['module_actuel'])] = self.pool.stats['serv_modules'].get(str(dico_modifs['module_actuel']), 0) + 1
        self.update_data()
        # tout s'est bien passé (on peut toujours rêver)
        return 1, config.u('ok')

    def _cree_arbo_serveur(self,new_module=None,new_variante=None):
        """crée l'arborescence zephir d'un nouveau serveur"""
        # création des différents répertoires du serveur sur zéphir
        edit_mode = False
        if new_module is not None:
            edit_mode = True
            module = new_module
        else:
            module = self.id_mod
        if new_variante is not None:
            edit_mode = True
            variante = new_variante
        else:
            variante = self.id_var
        serveur_dir = self.confdir

        if not edit_mode:
            try:
                os.makedirs(serveur_dir)
            except:
                return 0, """erreur de création du répertoire du serveur"""
            # création du répertoire de stockage des fichiers spécifiques au module
            try:
                os.makedirs(serveur_dir+os.sep+'fichiers_zephir')
                os.makedirs(serveur_dir+os.sep+'fichiers_perso')
                os.makedirs(serveur_dir+os.sep+'dicos')
                os.makedirs(serveur_dir+os.sep+'patchs')
                os.makedirs(serveur_dir+os.sep+'uucp')
            except:
                shutil.rmtree(serveur_dir)
                return 0, """erreur de création du répertoire des fichiers spécifiques au module"""

        if (not edit_mode) or new_module:
            # création du lien vers le dictionnaire du module
            if os.path.isdir(os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/dicos'):
                # module Eole2
                if not edit_mode:
                    os.makedirs(serveur_dir+os.sep+'dicos/local')
                retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/dicos '+serveur_dir+'/dicos/module')
                retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/module.eol '+serveur_dir+'/module.eol')
            else:
                # module Eole1
                retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/dictionnaire'+' '+serveur_dir+'/dictionnaire')
            if retour != 0:
                shutil.rmtree(serveur_dir)
                return 0, """erreur de lien sur le dictionnaire"""

        if (not edit_mode) or new_variante:
            # création des liens vers les patchs et dicos des variantes (ou du module si pas de variante)
            # dico.eol
            retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/variantes/'+str(variante)+os.sep+'dico.eol'+' '+serveur_dir+os.sep+'dico.eol')
            if retour == 0:
                retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/variantes/'+str(variante)+os.sep+'droits_zephir'+' '+serveur_dir+os.sep+'droits_variante')
                # patchs
                if retour == 0:
                    retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/variantes/'+str(variante)+'/patchs'+' '+serveur_dir+'/patchs/variante')
                    # dicos
                    if retour == 0:
                        retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/variantes/'+str(variante)+'/dicos'+' '+serveur_dir+'/dicos/variante')
                        # fichiers additionnels (définis dans les dictionnaires locaux)
                        if retour == 0:
                            retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/variantes/'+str(variante)+'/fichiers_perso'+' '+serveur_dir+'/fichiers_perso/variante')
                            # fichiers spécifiques au module (définis dans le dictionnaire zephir)
                            if retour == 0:
                                retour = os.system('ln -nsf '+os.path.abspath(config.PATH_MODULES)+os.sep+str(module)+'/variantes/'+str(variante)+'/fichiers_zephir'+' '+serveur_dir+'/fichiers_zephir/variante')

        if retour != 0:
            shutil.rmtree(serveur_dir)
            return 0, """erreur de création des liens vers la variante"""
        else:
            return 1, 'ok'


    def update_params(self,params_ori):
        # le champ params n'est pas initialisé
        if self.version == 'creole1':
            migration_ok = -1
        else:
            # pour les serveurs NG, on considère que la migration est déjà faite
            migration_ok = 1
        params = {'agents':1
                  ,'dico_ok':0
                  ,'config_ok':0
                  ,'migration_ok':migration_ok
                  ,'cle_ok':0
                  ,'new_key':(0,'')
                  #,'uucp_transfert':0
                  #,'uucp_cmd':0
                  ,'maj_ok':[-2,'']
                  ,'reboot_ok':[-2,'']
                  ,'service_restart_ok':[-2,'']
                  ,'query_maj':[-2,'']
                  ,'reconfigure_ok':[-2,'']
                  ,'configure_ok':[-2,'']
                  ,'sauvegarde_ok':[-2,'']
                  ,'upgrade_ok':[-2,'']
                  ,'perso_ok':[-2,'']
                  ,'lock_ok':[1,'']
                  ,'timeout':[1,'']
                  ,'md5s':[-1,'']
                  ,'last_log':'journal vide'
                }
        try:
            params['agents'] = params_ori['agents']
        except:
            # le serveur n'a pas encore d'info sur les agents
            pass
        # présence de zephir.eol et clé d'enregistrement
        if os.path.isfile(os.path.join(self.confdir,'zephir.eol')):
            params['config_ok']=1
        if os.path.isfile(os.path.join(self.confdir,'cle_publique')):
            params['cle_ok']=1
        if os.path.isfile(os.path.join(self.confdir,'new_key.pub')):
            if os.path.isfile(os.path.join(self.confdir,'new_addr')):
                params['new_key'] = (2,file(os.path.join(self.confdir,'new_addr')).read().strip())
            elif os.path.isfile(os.path.join(self.confdir,'new_addr_ok')):
                params['new_key'] = (3,file(os.path.join(self.confdir,'new_addr_ok')).read().strip())
            else:
                params['new_key'] = (1,'')
        # vérification des logs pour maj et reconfigure
        query = """select type,etat,date,message from last_log_serveur where id_serveur = %s order by date desc,id desc limit 1"""
        cu = cx_pool.create()
        try:
            cu.execute(query, (int(self.id_s),))
            data = cu.fetchone()
            # on stocke la date du dernier log remonté
            if data is not None:
                params['last_log']=str(data[2])
            query = """select type,etat,date,message from last_log_serveur where id_serveur = %s and type != 'SURVEILLANCE' order by date desc,id desc"""
            cu.execute(query, (int(self.id_s),))
            data = cu.fetchall()
            cx_pool.close(cu)
            # on parse les logs
            not_found = ['maj','reconfigure','configure','lock','sauvegarde','reboot','service_restart','upgrade','perso']
            for log in data:
                # on regarde si le serveur n'a pas remonté un bloquage des tâches zephir
                # (pb : procédure pour remonter la résolution du problème)
                if log[0] == 'LOCK':
                    if 'lock' in not_found:
                        not_found.remove('lock')
                        if log[1] == 1:
                            params['lock_ok'] = [2,str(log[2]),log[3]]
                else:
                    for type_log in ['maj','configure','reconfigure','sauvegarde','reboot','service_restart','upgrade','perso']:
                        # état des actions
                        if log[0] == type_log.upper() and int(log[1]) != -2:
                            # on ne stocke que le dernier etat de chaque action
                            if type_log in not_found:
                                not_found.remove(type_log)
                                if log[1] == -1:
                                    params['%s_ok' % type_log] = [2,str(log[2]),log[3]]
                                elif log[1] > 0:
                                    params['%s_ok' % type_log] = [0,str(log[2]),log[3]]
                                elif log[1] == 0:
                                    params['%s_ok' % type_log] = [1,str(log[2]),log[3]]
                        break
                if not_found == []:
                    break
        except:
            traceback.print_exc()
            cx_pool.close(cu)
        return params

    def maj_params(self, param_updates):
        """modifie le champ param d'un serveur
        @param_updates : dictionnaire contenant les modifications
        """
        params_ori = self.get_params()
        params_ori.update(param_updates)
        self.save_params(params_ori)

    def save_params(self, params):
        query = """update serveurs set params=E%s where id=%s"""
        sql_params = (str(params), int(self.id_s))
        cu = cx_pool.create()
        try:
            cu.execute(query, sql_params)
            cx_pool.commit(cu)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)

    def check_md5conf(self):
        # création de la liste des fichiers présents sur zephir
        modifs = []
        fics=[]
        files=md5files[self.module_version]
        # cas particulier : zephir.eol
        # creole 1 : suppression des lignes sans variables
        # creole 2 : valeurs précédentes ignorées
        fic_ori = os.path.join(self.confdir,'zephir.eol')
        fic_var = os.path.join(self.confdir,'variables.eol')
        if os.path.isfile(fic_ori):
            f_var = file(fic_var,'w')
            for line in file(fic_ori).readlines():
                if self.module_version == 1:
                    if "@@" in line:
                        f_var.write(line)
                else:
                    if not "valprec =" in line:
                        f_var.write(line)
            f_var.close()
        for src, dst, pattern in files:
            if os.path.isdir(os.path.join(self.confdir,dst)):
                for fic in os.listdir(os.path.join(self.confdir,dst)):
                    if os.path.isfile(os.path.join(self.confdir,dst,fic)):
                        if pattern == None or fic.endswith(pattern):
                            fics.append(os.path.join(dst,fic))
            else:
                fics.append(dst)
        md5s = []
        md5file = os.path.join(os.path.abspath(config.PATH_ZEPHIR),'data','config%s.md5' % self.id_s)
        if os.path.isfile(md5file):
            # liste des fichiers remontés par le serveur
            data = file(md5file).read().strip().split('\n')
            # comparaison
            for line in data:
                fic = line.split()[1]
                if fic not in fics:
                    modifs.append("absent sur zephir : %s" % fic)
                else:
                    md5s.append(fic)
            for fic in fics:
                if fic not in md5s:
                    # erreur : fichier non présent sur la machine distante
                    modifs.append("non envoyé : %s" % fic)
            # test du contenu des fichiers
            cmd_md5 = getProcessOutputAndValue("/usr/bin/md5sum",
                                        args = ["-c",md5file],
                                        path = self.confdir,
                                        env = {'LC_ALL': 'C'})
            cmd_md5.addCallback(self._check_md5_res, md5s, modifs)
            cmd_md5.addErrback(self._check_md5_res, md5s, modifs)
        else:
            # pas d'infos md5 pour ce serveur
            self.maj_params({'md5s':[-1,""]})
            self.edit_serveur({'md5s':-1})

    def _check_md5_res(self, result, md5s, modifs):
        # on effectue une vérification md5 des fichiers de configuration si disponible
        out, err, code = result
        self.edit_serveur({'md5s':1})
        self.maj_params({'md5s':[1,""]})
        for fic in md5s:
            if not "%s: OK" % fic in out.split('\n'):
                if fic == 'variables.eol':
                    # cas particulier des variables creole 1, on indique une modification sur zephir.eol
                    modifs.append("contenu modifié : zephir.eol")
                else:
                    modifs.append("contenu modifié : %s" % fic)
                md5_ok = False
        if modifs != []:
            self.edit_serveur({'md5s':0})
            self.maj_params({'md5s':[0,";".join(modifs).strip()]})

    def check_maj_status(self, show_installed = False, debnames = []):
        """vérifie l'état de mise à jour d'un serveur
        """
        if self.version == 'creole1':
            return []
        # liste des paquets remontés par le serveur
        if type(debnames) != list:
            debnames = [debnames]
        liste_pkgs = []
        # on regarde le type de mise à jour configurée (minimale/complète) et le serveur de mise à jour
        dico = self.get_config()
        type_maj = dico.get_value('type_maj')[0]
        serv_maj = dico.get_value('serveur_maj')[0]
        version = self.module_version
        pkg_file = os.path.join(os.path.abspath(config.PATH_ZEPHIR),'data','packages%s.list' % self.id_s)
        # on compare les paquets installés avec ceux disponibles sur le serveur de maj
        liste_pkgs, maj_infos = maj_checker.check_packages(self.maj_infos, pkg_file, debnames, version, serv_maj, type_maj, show_installed)
        if maj_infos is not None and not debnames and not show_installed and maj_infos != self.maj_infos:
            # la liste des paquets non à jour a changé, on met à jour l'état du serveur
            try:
                nb_pack = len(maj_infos[0])
                if len(maj_infos[0]) > 0:
                    self.edit_serveur({'maj':len(maj_infos[0])})
                    self.maj_params({'query_maj':[len(maj_infos[0]),time.ctime()]})
                else:
                    self.edit_serveur({'maj':0})
                    self.maj_params({'query_maj':[0,time.ctime()]})
                self.maj_infos = maj_infos
            except:
                traceback.print_exc()
                pass
        return liste_pkgs

    def check_min_version(self, debname, min_version):
        # recherche de la version actuelle du paquet sur ce serveur
        pkg_file = os.path.join(os.path.abspath(config.PATH_ZEPHIR),'data','packages%s.list' % self.id_s)
        if os.path.isfile(pkg_file):
            for line in file(pkg_file):
                pak_name, pak_version = line.split()
                if pak_name == debname:
                    return maj_checker.check_min_version(pak_version, min_version)
        # information non trouvée, on retourne None
        return None

    def regen_key(self, new_addr=None):
        """prépare une nouvelle clé ssh pour la connexion uucp à zephir
        """
        new_key = os.path.join(self.confdir, 'new_key')
        addr_file = os.path.join(self.confdir, 'new_addr')
        if os.path.exists(new_key):
            os.unlink(new_key)
        if os.path.exists("%s.pub" % new_key):
            os.unlink("%s.pub" % new_key)
        for f_addr in [addr_file, '%s_ok' % addr_file]:
            if os.path.exists(f_addr):
                os.unlink(f_addr)
        zephir_addr =  new_addr or config.ADRESSE_ZEPHIR
        res = os.system("""/usr/bin/ssh-keygen -N "" -b 1024 -t rsa -f %s -C uucp@%s""" % (new_key, zephir_addr))
        # si nouvelle adresse à envoyer, on crée un fichier new_addr
        if new_addr:
            f_adr = open(addr_file, 'w')
            f_adr.write(new_addr)
            f_adr.close()
        return res

    def get_key(self, old_key, new_key, confirm_ip):
        """envoie une nouvelle clé à un serveur.
        les 2 clés publiques (ancienne et nouvelles) doivent être passées en paramètre.
        """
        try:
            old_key_path = os.path.join(self.confdir, 'cle_publique')
            new_key_path = os.path.join(self.confdir, 'new_key.pub')
            new_key_priv_path = os.path.join(self.confdir, 'new_key')
            new_addr_path = os.path.join(self.confdir, 'new_addr')
            old_data = file(old_key_path).read().strip()
            old_data = base64.encodestring(old_data[old_data.index('ssh-rsa'):])
            if old_data == old_key:
                # la clé correspond, on renvoie les nouvelles clés
                data_keys = []
                new_data = file(new_key_path).read().strip()
                new_data = base64.encodestring(new_data[new_data.index('ssh-rsa'):])
                if new_data != new_key:
                    return 0, "nouvelle clé publique non reconnue"
                if confirm_ip == True:
                    if os.path.isfile(new_addr_path):
                        os.rename(new_addr_path, '%s_ok' % new_addr_path)
                        return 1, ''
                    else:
                        return 0, "changement d'adresse non détecté sur zephir"
                new_data_priv = base64.encodestring(file(new_key_priv_path).read().strip())
                os.unlink(new_key_priv_path)
                return 1, new_data_priv
            else:
                return 0, "ancienne clé publique non reconnue"
        except:
            return 0, "erreur de récupération des données"

    def get_last_contact(self):
        """renvoie la date de dernier contact avec le serveur
        """
        cu = cx_pool.create()
        try:
            cu.execute("select last_contact from serveurs where id=%s", (int(self.id_s),))
            data = cu.fetchone()
            cx_pool.close(cu)
        except:
            traceback.print_exc()
            cx_pool.close(cu)
            return 0.0
        return float(data[0])

    def migrate_config(self,module_dest,variante_dest,mode="migration"):
        """crée un dictionnaire creole2 et importe les valeurs du serveur"""
        assert self.version == 'creole1'
        assert int(module_dest) >= 1
        assert int(variante_dest) >= 1
        # chemin vers les dictionnaires du module
        creoledirs = [os.path.join(os.path.abspath(config.PATH_MODULES),str(module_dest),'dicos')]
        creoledirs.append(os.path.join(os.path.abspath(config.PATH_MODULES),str(module_dest),'variantes',str(variante_dest),'dicos'))
        # création du dictionnaire
        dico = ZephirDict(creoledirs, self.confdir, mode)
        # on stocke la variable choisie dans le répertoire de conf
        f_var = open(os.path.join(self.confdir, 'variante_migration'), 'w')
        f_var.write(str(variante_dest))
        f_var.close()
        return dico

    def migrate_data(self, check=False):
        """fonction de récupération des données d'un serveur migré
        check : vérifie seulement si des données sont migrables sur ce serveur
        """
        # fichiers à migrer : fichiers par défaut
        migration_files = {}
        module = self.module[:self.module.rindex('-')]
        if module in config.migration_files:
            migration_files.update(config.migration_files[module])
        # définitions supplémentaires
        if module in migration_perso:
            for section in migration_perso[module].keys():
                migration_files[section].extend(migration_perso[module][section])
        dir_serv = os.path.join(self.confdir,'fichiers_zephir')
        dir_bak = os.path.join(self.confdir+'-backup','fichiers_zephir')
        # mode check : indique si des fichiers sont gérés pour ce serveur
        if check == True:
            if len(migration_files.keys()) > 0 and os.path.isdir(dir_bak):
                return 1, True
            else:
                return 1, False
        # liste des fichiers à copier
        if not len(migration_files.keys()) > 0:
            return 1, config.u("""pas de fichiers à migrer""")
        if not os.path.isdir(dir_bak):
            return 0, config.u("répertoire backup de migration non trouvé")
        not_found = []
        errors = []
        # copie des fichiers avec conversion de l'encoding si besoin
        for src, dst, convert in migration_files['files']:
            fic_src = os.path.join(dir_bak,src)
            fic_dst = os.path.join(dir_serv,dst)
            if os.path.isdir(fic_src):
                # répertoire
                os.path.walk(fic_src, self.tree_copy, (fic_src, fic_dst, convert, errors, migration_files))
            elif os.path.isfile(fic_src):
                self.copy_fic(fic_src, fic_dst, convert, errors)
            else:
                not_found.append(src)
        # application des droits
        droits_zephir = []
        dests=[]
        for dest, options, user, group, mode in migration_files['rights']:
            if dest.endswith("*"):
                # cas "repertoire/*" : on recherche tous les fichiers
                dests = glob(os.path.join(dir_serv,dest))
            else:
                if os.path.exists(os.path.join(dir_serv,dest)):
                    dests = [os.path.join(dir_serv,dest)]
            for fic_dest in dests:
                if os.path.basename(fic_dest) not in migration_files['exclude']:
                    cmd_rights = 'chmod %s %s %s;chown %s %s.%s %s' % (options, mode, fic_dest, options, user, group, fic_dest)
                    res = os.system(cmd_rights)
                    if res != 0:
                        errors.append('application des droits impossible : %s' % fic_dest)
                    else:
                        # ajout de la ligne pour droits_zephir
                        droits_zephir.append("%s#%s#%s#%s#%s" % (fic_dest[fic_dest.index('fichiers_zephir'):],mode,user,group,'-R' in options))

        # création du fichier droits_zephir
        fic_droits = open(os.path.join(self.confdir, 'droits_zephir'),'w')
        fic_droits.write('\n'.join(droits_zephir))
        fic_droits.close()
        # mise en place du fichier fichiers_zephir (copie de fichiers_module + destinations personnalisées)
        # on utilise toujours le fichier de la version 2.0 pour une migration
        module_ng_orig = '%s-2.0' % self.module[:self.module.rindex('-')]
        module_file = '/usr/share/zephir/dictionnaires/fichiers_%s' % module_ng_orig
        serveur_file = os.path.join(self.confdir,'fichiers_zephir','fichiers_zephir')
        if os.path.isfile(module_file):
            # lecture du fichier du module
            data = file(module_file).read().strip()
        if module in destinations_perso:
            data += "\n# fichiers ajoutés lors de la migration Eole 1 vers Eole NG"
            for dest in destinations_perso[module]:
                data += "\n%s" % dest

        # écriture en utilisant les templates définis dans config.py
        final_data = "%s\n%s\n%%%%\n%s" % (config.FILE_SECTION,data,config.RPM_SECTION)

        # écriture du fichier final
        dests = file(serveur_file,'w')
        dests.write(final_data)
        dests.close()

        return 1, config.u((not_found,errors))

    def variante_migration(self):
        path_var = os.path.join(self.confdir, 'variante_migration')
        if os.path.isfile(path_var):
            return open(path_var).read().strip()
        else:
            return ''

    def add_replication(self, rne, content):
        # chemin de sauvegarde de l'archive contenant la conf de réplication
        repl_dir = os.path.join(self.confdir, 'replication')
        if not os.path.exists(repl_dir):
            os.makedirs(repl_dir)
        script = os.path.join(repl_dir, 'replication-%s.conf' % rne)
        # sauvegarde du script
        try:
            bin_file = StringIO()
            data = base64.decodestring(content)
            fd = open(script,'wb')
            # sauvegarde du fichier
            bin_file.write(data)
            bin_file.seek(0)
            fd.write(bin_file.read())
            fd.close()
        except:
            traceback.print_exc()
            return 0, config.u("erreur de l'écriture du fichier %s" % script)
        return 1, "OK"

    def del_replication(self, filename):
        repl_dir = os.path.join(self.confdir, 'replication')
        try:
            assert os.path.isfile(os.path.join(repl_dir, filename))
            os.unlink(os.path.join(repl_dir, filename))
            # création d'un fichier .modified pour indiquer que la conf est modifiée manuellement
            file(os.path.join(self.confdir, 'replication', '.modified'), 'w').close()
        except:
            return 0, config.u("Erreur lors de la suppression du fichier %s" % filename)
        return 1, "OK"

    def check_replication(self):
        # vérification du répertoire
        if not os.path.isdir(os.path.join(self.confdir, 'replication')):
            repl_state = 0
        elif os.path.isfile(os.path.join(self.confdir, 'replication', '.modified')):
            repl_state = 2
        else:
            repl_state = 1
        return 1, repl_state

    def get_replication(self):
        # définition du répertoire du serveur
        conf_repl = []
        # lecture de la liste des fichiers enregistrés
        if os.path.isdir(os.path.join(self.confdir, 'replication')):
            search_path = os.path.join(self.confdir, 'replication', 'replication-*.conf')
            try:
                for conf_file in glob(search_path):
                    conf_repl.append(os.path.basename(conf_file))
            except:
                return 0, config.u("Erreur lors de la recherche des configurations de réplication")
        return 1, config.u(conf_repl)

    def tree_copy(self, args, dirname, fnames):
        fic_src, fic_dst, convert, errors, migration_files = args
        if not os.path.isdir(fic_dst):
            os.makedirs(fic_dst)
        for fic in fnames:
            if fic not in migration_files['exclude']:
                src = os.path.join(dirname, fic)
                dst = src.replace(fic_src, fic_dst)
                if os.path.exists(src):
                    self.copy_fic(src, dst, convert, errors)

    def copy_fic(self, src, dst, convert, errors):
        """copie du fichier src sur dst
        convert: si True, convertit le fichier d'ISO-8859-1 vers UTF-8
        """
        if convert == True:
            cmd = '/usr/bin/iconv -f ISO-8859-1 -t UTF-8 -o "%s" "%s"' % (dst, src)
        else:
            cmd = '/bin/cp -f "%s" "%s"' % (src, dst)
        res = os.system(cmd)
        if res != 0:
            log.msg("erreur lors de la migration des données, la commande suivante a échoué :", cmd)
            errors.append(src)


class ServeurPool(dict):
    """dictionnaire des serveurs existants gérant les autorisations d'accès
    """

    def __init__(self):

        self._groupes_serveurs = {}
        self._restrictions = {}
        self.stats = {}
        # self.update_groupes()
        # dictionnaire des variantes et leur module
        self._mod_var = self.get_mod_var()
        try:
            self.update_auths()
        except:
            # pas encore de table de restriction sur cette version
            pass

    def update_groupes(self):
        cu = cx_pool.create()
        # récupération des groupes de serveurs
        try:
            cu.execute("select id, libelle, serveurs from groupes_serveurs")
            data = cu.fetchall()
            need_commit = False
            for groupe in data:
                id = int(groupe[0])
                libelle = str(groupe[1])
                serveurs = []
                # suppression des éventuels serveurs inexistants
                update_gr = False
                for serveur in eval(groupe[2]):
                    if serveur in self:
                        serveurs.append(serveur)
                    else:
                        update_gr = True
                if update_gr:
                    # des serveurs ne sont plus valides, on les supprime du groupe dans la base de données
                    cu.execute("update groupes_serveurs set serveurs=%s where id=%s", (str(serveurs), int(id)))
                    need_commit = True
                timestamp = time.time()
                date_creat = timestamp
                if self._groupes_serveurs.has_key(id):
                    date_creat = self._groupes_serveurs[id][2][0]
                self._groupes_serveurs[id] = [libelle, serveurs, [date_creat, timestamp]]
            if need_commit:
                cx_pool.commit(cu)
            else:
                cx_pool.close(cu)
        except:
            traceback.print_exc()
            cx_pool.close(cu)

    def get_mod_var(self):
        """récupére les variantes et leur module"""
        cu = cx_pool.create()
        try:
            cu.execute("select id, module from variantes")
            data = cu.fetchall()
            cx_pool.close(cu)
        except:
            data = []
            traceback.print_exc()
            cx_pool.close(cu)
        return dict(data)

    def update_auths(self, user=None):
        params = []
        if user != None:
            user_clause = " where login=%s"
            params.append(user)
        else:
            user_clause = ""

        cu = cx_pool.create()
        try:
            # récupération des types de contraintes spécifiées
            cu.execute("select login, id_res, type_res from restrictions %s" % user_clause, params)
            data = cu.fetchall()
            for restriction in data:
                login, id_res, type_res = restriction
                if not self._restrictions.has_key(login):
                    self._restrictions[login] = {}
                if self._restrictions[login].has_key(type_res):
                    self._restrictions[login][type_res].append(id_res)
                else:
                    self._restrictions[login][type_res] = [id_res]
            cx_pool.close(cu)
        except:
            traceback.print_exc()
            cx_pool.close(cu)

    def add_restriction(self, credential, type_res, id_res):
        """ajoute une restriction pour un utilisateur et une ressource donnés
        type_res : nature des objets à restreindre
        id_res ; identifiant de l'objet autorisé
        les différents types reconnus sont : 'rne, group, id_mod, id_var' et les attributs
        de la classe serveur (id_s, module, version, ...)
        """
        # ajout dans la base
        id_res = str(id_res)
        try:
            assert id_res in self._restrictions[credential][type_res]
        except:
            cu = cx_pool.create()
            # validation de l'existence
            libelle, table, field = config.type_res_label[type_res]
            if field in ('rne','libelle'):
                # données de type str
                query = "select * from %s where %s ilike %%s" % (table, field)
                params = (id_res,)
            else:
                # type entier
                query = "select * from %s where %s = %%s" % (table, field)
                params = (int(id_res),)
            try:
                cu.execute(query, params)
                # pas de données correspondante trouvée
                if cu.fetchone() == None:
                    raise ValueError('identifiant de ressource inconnu')
            except Exception, e:
                cx_pool.close(cu)
                return False
            try:
                params = (credential, type_res, id_res)
                cu.execute("""insert into restrictions (login, type_res, id_res) values (E%s,E%s,E%s)""", params)
            except:
                # erreur d'insertion dans la base
                cx_pool.rollback(cu)
                return False
            cx_pool.commit(cu)
            if not self._restrictions.has_key(credential):
                self._restrictions[credential] = {}
            if not self._restrictions[credential].has_key(type_res):
                self._restrictions[credential][type_res] = []
            self._restrictions[credential][type_res].append(id_res)
        else:
            # restriction déjà existante
            pass
        return True

    def get_restrictions(self, credential, type_res=None):
        """renvoie la liste des restrictions d'un utilisateur
        """
        if not self._restrictions.has_key(credential):
            return []
        if type_res == None:
            # toutes les restrictions : dictionnaire
            return self._restrictions[credential]
        else:
            # si restrictions d'un type précis, on renvoie une liste de ressources
            if self._restrictions[credential].has_key(type_res):
                return self._restrictions[credential][type_res]
            else:
                return []

    def del_restriction(self, credential, type_res, id_res):
        """enleve une restriction d'un utilisateur
        """
        id_res = str(id_res)
        try:
            liste_res = self._restrictions[credential][type_res]
            assert id_res in liste_res
        except:
            # restriction non existante
            return False
        # suppression dans la base
        cu = cx_pool.create()
        try:
            cu.execute("""delete from restrictions where login=%s and type_res=%s and id_res=%s""", (credential, type_res, id_res))
            cx_pool.commit(cu)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)
            return False
        # supression en mémoire
        self._restrictions[credential][type_res].remove(id_res)
        if self._restrictions[credential][type_res] == []:
            del(self._restrictions[credential][type_res])
        return True

    def get_file_perms(self, data_dir, filepath=""):
        """renvoie les informations de permissions associées à un fichier
        """
        # lecture des informations sur le fichier si disponibles
        permsfile = data_dir+os.sep+'droits_zephir'
        if os.path.isfile(permsfile):
            f_rights = file(permsfile)
            data = f_rights.read().strip().split('\n')
            f_rights.close()
        else:
            data = []
        mode = user = group = ""
        recursive = False
        if filepath != "":
            result = {filepath:[mode, user, group, recursive]}
        else:
            result = {}
        for line in data:
            if line.startswith(filepath+'#') or filepath == "":
                try:
                    filename, mode, user, group, recursive = line.split('#')
                    if recursive != '':
                        recursive = eval(recursive)
                    result[filename] = [mode, user, group, recursive]
                    if filepath != "":
                        break
                except:
                    # fichier vide ?
                    pass
        return result

    def set_file_perms(self, rights, data_dir):
        """enregistre les informations de permissions associées à un(des) fichier(s)
        @param data_dir: chemin ou trouver le fichier droits_zephir
        """
        # lecture des données existantes
        permsfile = data_dir+os.sep+'droits_zephir'
        data = self.get_file_perms(data_dir)
        data.update(rights)
        # stockage des informations
        lines = []
        for filepath in data.keys():
            # vérification des données
            lines.append("%s#%s#%s#%s#%s" % (filepath, data[filepath][0], data[filepath][1], data[filepath][2],data[filepath][3]))
        # écriture des informations
        f_rights = file(permsfile,'w')
        f_rights.write('\n'.join(lines))
        f_rights.close()
        return True

    def del_file_perms(self, data_dir, filepath="", recurse=False):
        """supprime les informations de permissions associées à un fichier (ou tous)
        """
        # lecture des informations sur le fichier si disponibles
        permsfile = data_dir+os.sep+'droits_zephir'
        if not os.path.exists(permsfile):
            # pas de permissions exitantes ou mauvais chemin
            return False
        if filepath != "":
            f_rights = file(permsfile)
            data = f_rights.read().strip().split('\n')
            f_rights.close()
            if recurse == True:
                search_pattern = filepath
            else:
                search_pattern = filepath + '#'
            for line in data:
                if line.startswith(search_pattern):
                    data.remove(line)
        else:
            data = []
        # écriture des informations
        try:
            f_rights = file(permsfile,'w')
            f_rights.write('\n'.join(data))
            f_rights.close()
        except:
            return False
        return True

    def check_serv_credential(self, credential, key):
        # vérification des restrictions sur les attributs du serveur
        # (id_s, rne, id_mod, variante)
        serv = dict.get(self,key)
        if self._restrictions.has_key(credential):
            for type_res, res in self._restrictions[credential].items():
                if res == []:
                    # ce cas ne devrait pas arriver car si on passe toujours par del_restriction
                    continue
                res_ok = True
                # si la ressource n'est pas autorisée
                # cas particuliers : rne et groupes
                if type_res == 'rne':
                    rne_ok = False
                    for rne in res:
                        # cas des rne avec wildcard (%)
                        if rne.endswith('%'):
                            if serv.rne.startswith(rne.split('%')[0]):
                                # ce rne correspond à l'expression
                                rne_ok = True
                                break
                    if not rne_ok:
                        if not serv.rne in res:
                            res_ok = False
                elif type_res == 'groupe':
                    # test d'appartenance à un groupe autorisé
                    serv_in_groupes = False
                    for groupe in res:
                        if serv.id_s in self._groupes_serveurs[int(groupe)][1]:
                            serv_in_groupes = True
                            break
                    if serv_in_groupes == False:
                        res_ok = False
                else:
                    # test attribut correspondant du serveur dans les valeurs autorisées
                    if not str(getattr(serv,type_res)) in [str(i) for i in res]:
                        res_ok = False
                if not res_ok:
                    # on interdit l'accès
                    raise ResourceAuthError("unauthorized resource  : %s" % key)

    def check_gr_credential(self, credential, groupe):
        # validation des droits d'accès à un groupe
        if self._restrictions.has_key(credential):
            if self._restrictions[credential].has_key('groupe'):
                if self._restrictions[credential]['groupe'] != []:
                    if str(groupe) not in self._restrictions[credential]['groupe']:
                        # accès refusé
                        # FIXME vérification suffisante ou regarder les serveurs ou rne du groupe ?
                        raise ResourceAuthError("unauthorized resource  : %s" % groupe)

    def check_mod_credential(self, credential, id_mod):
        # validation des droits d'accès à un module
        if self._restrictions.has_key(credential):
            if self._restrictions[credential].has_key('id_mod'):
                if self._restrictions[credential]['id_mod'] != []:
                    if str(id_mod) not in self._restrictions[credential]['id_mod']:
                        # accès refusé
                        raise ResourceAuthError("unauthorized module  : %s" % id_mod)

    def check_var_credential(self, credential, id_var):
        # validation des droits d'accès à une variante
        if self._restrictions.has_key(credential):
            try:
                var_mod = str(self._mod_var[int(id_var)])
            except:
                # on met à jour si la variante n'est pas trouvée
                self._mod_var = self.get_mod_var()
                var_mod = str(self._mod_var[int(id_var)])
            # on commence par vérifier le module auquel la variante appartient
            if self._restrictions[credential].has_key('id_mod'):
                if self._restrictions[credential]['id_mod'] != []:
                    if var_mod not in self._restrictions[credential]['id_mod']:
                        raise ResourceAuthError("accès au module %s interdit" % var_mod)
            # vérification de la variante elle-même
            if self._restrictions[credential].has_key('id_var'):
                if self._restrictions[credential]['id_var'] != []:
                    if str(id_var) not in self._restrictions[credential]['id_var']:
                        # accès refusé
                        raise ResourceAuthError("accès à la variante %s interdit" % id_var)

    def check_etab_credential(self, credential, rne):
        # validation des droits d'accès à un établissement
        rne_ok = True
        if self._restrictions.has_key(credential):
            if self._restrictions[credential].has_key('rne'):
                if self._restrictions[credential]['rne'] != []:
                    if str(rne) not in self._restrictions[credential]['rne']:
                        rne_ok = False
                        for etab in self._restrictions[credential]['rne']:
                            if etab.endswith('%'):
                                if str(rne).startswith(etab.split('%')[0]):
                                    rne_ok = True
        if not rne_ok:
            # accès refusé
            raise ResourceAuthError("unauthorized module  : %s" % rne)

    def get(self, credential, key):
        if key in self.keys():
            self.check_serv_credential(credential, key)
        else:
            raise KeyError, "serveur inexistant"
        return dict.get(self,key)

    def get_stats(self):
        self.stats['nb_serv'] = len(self)
        return self.stats

    def get_alertes(self, credential):
        """renvoie la liste des serveurs en alerte
        """
        servs=[]
        for id_serv in self.keys():
            try:
                serv = self.get(credential, int(id_serv))
            except ResourceAuthError:
                # accès à ce serveur non autorisé
                continue
            if serv.get_status() not in [1, -1]:
                servs.append(serv)
        return servs

    def get_migration_status(self, credential):
        """renvoie la liste des serveurs en cours de migration/non migrés
        """
        servs=[[],[]]
        for id_serv in self.keys():
            try:
                serv = self.get(credential, int(id_serv))
            except ResourceAuthError:
                # accès à ce serveur non autorisé
                continue
            params = serv.get_params()
            if params['migration_ok'] == 0:
                servs[0].append(serv)
            elif params['migration_ok'] == -1:
                servs[1].append(serv)
        return servs

    def check_serveurs(self, credential, serveurs, last_check=None):
        """vérifie si des serveurs ont été modifiés
        @param serveurs : dictionnaires {idserveur:timestamp}
        @param last_check : Si != None, on renvoie les serveurs créés après cette date (timestamp)
        @return : liste d'id de serveurs
        """
        modifs = {}
        serv_pool = []
        for id_serv in [str(cle) for cle in self.keys()]:
            try:
                serv = self.get(credential, int(id_serv))
            except ResourceAuthError:
                # serveur non existant ou non accessible
                log.msg("server %s not authorized" % id_serv)
                if serveurs.has_key(id_serv):
                    modifs[id_serv] = "deleted"
            else:
                serv_pool.append(str(serv.id_s))
                if serveurs.has_key(id_serv):
                    if serv.modified > serveurs[id_serv]:
                        # serveur modifié
                        modifs[id_serv] = serv.modified
                elif last_check != None:
                    if serv.created > last_check:
                        # si le serveur a été créé après la dernière mise à jour du client
                        # on l'ajoute comme nouveau serveur
                        modifs[id_serv] = serv.modified

        for id_serv in serveurs.keys():
            if id_serv not in serv_pool:
                modifs[id_serv] = "deleted"

        return modifs

    def check_groupes(self, credential, groupes, last_check=None):
        """vérifie si des groupes ont été modifiés
        @param groupes : dictionnaires {idgroupe:timestamp}
        @param last_check : Si != None, on renvoie les serveurs créés après cette date (timestamp)
        @return : dictionnaire  {id_gr:[libelle, serveurs, [date_creat, date_modif]]}
        """
        modifs = {}

        for id_gr, data in self._groupes_serveurs.items():
            try:
                group = self.get_groupes(credential, id_gr)
            except KeyError, ResourceAuthError:
                # groupe non existant ou non accessible
                log.msg("group %s no longer exists or not authorized" % id_gr)
                modifs[str(id_gr)] = "deleted"
            else:
                # la clé des dictionnaires doit être une chaine pour xmlrpc
                id_gr = str(id_gr)
                if groupes.has_key(id_gr):
                    if data[2][1] > groupes[id_gr]:
                        # groupe modifié
                        modifs[id_gr] = data
                elif last_check != None:
                    if data[2][0] > last_check:
                        # nouveau groupe
                        modifs[id_gr] = data
        return modifs

    def add_groupe(self, credential, libelle, serveurs):
        query = """insert into groupes_serveurs (libelle,serveurs) values (E%s, %s)"""
        params = (libelle, str(serveurs))
        cu = cx_pool.create()
        try:
            cu.execute(query, params)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)
            return False
        cx_pool.commit(cu)
        self.update_groupes()
        # ajout automatique du groupe pour l'utilisateur l'ayant créé si il a des restrictions de groupes
        if self._restrictions.has_key(credential):
            if self._restrictions[credential].has_key('groupe'):
                # on recherche l'id du groupe ajouté
                for id_gr in self._groupes_serveurs.keys():
                    if self._groupes_serveurs[id_gr][0] == libelle:
                        self.add_restriction(credential,'groupe',str(id_gr))
        return True

    def get_groupes(self, credential, id_groupe=None, alertes=False):
        if id_groupe != None:
            groupes = [int(id_groupe)]
        else:
            groupes = self._groupes_serveurs.keys()
        # vérification des restrictions sur les groupes
        res = []
        for groupe in groupes:
            try:
                self.check_gr_credential(credential, groupe)
            except:
                # groupe non autorisé, on passe au suivant
                continue
            libelle, serveurs, timestamps = self._groupes_serveurs[groupe]
            # sinon on ajoute le groupe à la liste
            res.append([int(groupe), libelle, serveurs])
        return res

    def edit_groupe(self, credential, id_groupe, libelle, serveurs):
        # vérification des droits sur le groupe
        id_groupe = int(id_groupe)
        self.check_gr_credential(credential, id_groupe)
        # mise à jour de la base
        cu = cx_pool.create()
        query = """update groupes_serveurs set libelle=E%s, serveurs=%s where id=%s"""
        params = (libelle,str(serveurs),int(id_groupe))
        try:
            cu.execute(query, params)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)
            return False
        cx_pool.commit(cu)
        # mise à jour interne
        date_creat = self._groupes_serveurs[id_groupe][2][0]
        self._groupes_serveurs[id_groupe] = [libelle, serveurs, [date_creat,time.time()]]
        return True

    def del_groupe(self, credential, id_groupe):
        # vérification des droits sur le groupe
        id_groupe = int(id_groupe)
        self.check_gr_credential(credential, id_groupe)
        # on supprime le groupe des groupes surveillés par les utilisateurs
        rech_params = ('[%d]' % id_groupe, '[%%, %d, %%]' % id_groupe, '[%%, %d]' % id_groupe, '[%d, %%]' % id_groupe)
        cursor = cx_pool.create()
        try:
            cursor.execute("""select login,groupes from users where groupes like %s or groupes like %s or groupes like %s or groupes like %s""", rech_params)
            data=cursor.fetchall()
            # pour chaque utilisateur
            for user in data:
                # modification de la liste des groupes
                groupes = eval(user[1])
                if id_groupe in groupes:
                    groupes.remove(id_groupe)
                    sql_update = """update users set groupes=%s where login=E%s"""
                    params = (str(groupes), user[0])
                    cursor.execute(sql_update, params)
            # suppression du groupe dans la base
            cursor.execute("""delete from groupes_serveurs where id = %s""", (int(id_groupe),))
            cx_pool.commit(cursor)
        except:
            traceback.print_exc()
            cx_pool.rollback(cursor)
            return False
        # suppression du groupe en mémoire
        del(self._groupes_serveurs[id_groupe])
        return True


    def extend_groupe(self, credential, id_groupe, serveurs):
        # vérification des droits sur le groupe
        id_groupe = int(id_groupe)
        self.check_gr_credential(credential, id_groupe)
        # extension des serveurs du groupe
        # dans le dictionnaire interne
        for serv in serveurs:
            if serv not in self._groupes_serveurs[id_groupe][1]:
                self._groupes_serveurs[id_groupe][1].append(serv)
        self._groupes_serveurs[id_groupe][2] = [time.time(),time.time()]
        cu = cx_pool.create()
        # dans la base de données
        try:
            query = """update groupes_serveurs set serveurs=%s where id=%s"""
            params = (str(self._groupes_serveurs[id_groupe][1]), id_groupe)
            cu.execute(query, params)
            cx_pool.commit(cu)
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)

    def add_serveur(self, credential, rne, libelle, materiel, processeur, disque_dur, date_install, installateur, tel, remarques, module_initial, module_actuel, variante, timestamp_serveur, timeout):
        # on vérifie si l'accès au rne , module et variante est permis
        if self._restrictions.has_key(credential):
            if self._restrictions[credential].has_key('rne'):
                self.check_etab_credential(credential, rne)
            if self._restrictions[credential].has_key('id_mod'):
                if module_actuel not in self._restrictions[credential]['id_mod']:
                    raise ResourceAuthError("unauthorized resource %s" % module_actuel)
            if self._restrictions[credential].has_key('variante'):
                cu = cx_pool.create()
                cu.execute('select libelle from variantes where id=%s', (int(variante),))
                libel_var = cu.fetchone()[0]
                cx_pool.close(cu)
                # dans le cas d'une variante standard, on l'autorise toujours
                if libel_var != 'standard' and variante not in self._restrictions[credential]['variante']:
                    raise ResourceAuthError("unauthorized resource %s" % variante)
        # insertion dans la base de données
        cu = cx_pool.create()
        query = """insert into serveurs (rne,libelle,materiel,processeur,disque_dur,date_install,installateur,tel,remarques,module_initial,module_actuel,variante,timestamp,timeout) values (E%s,E%s,E%s,E%s,E%s,E%s,E%s,%s,E%s,%s,%s,%s,%s,%s)"""
        params = (rne, libelle, materiel, processeur, disque_dur, date_install, installateur, tel, remarques, int(module_initial), int(module_actuel), int(variante), timestamp_serveur, int(timeout))
        cu.execute(query, params)
        # récupération de l'identifiant attribué par la base
        query = """select id from serveurs where rne=E%s and libelle=E%s and date_install=E%s and installateur=E%s and remarques=E%s and module_actuel=%s and variante=%s and timestamp=%s"""
        params = (rne, libelle, date_install, installateur, remarques, int(module_actuel), int(variante), timestamp_serveur)
        cu.execute(query, params)
        id_serveur = int(cu.fetchone()[0])
        self[id_serveur] = Serveur(self, id_serveur,cu)
        self.stats['no_contact'].append(id_serveur)
        self.stats['serv_modules'][str(module_actuel)] = self.stats['serv_modules'].get(str(module_actuel),0) + 1
        self.stats['serv_variantes'][str(variante)] = self.stats['serv_variantes'].get(str(variante),0) + 1
        cx_pool.commit(cu)
        return self[id_serveur]

    def del_serveur(self, credential, id_serveur):
        """Supression d'un serveur de la base
        """
        # vérification des restrictions
        self.check_serv_credential(credential, id_serveur)
        cu = cx_pool.create()
        # on supprime le serveur dans la base
        try:
            cu.execute("""delete from serveurs where id=%s""", (int(id_serveur),))
            cx_pool.commit(cu)
            # supression interne
            self.stats['serv_modules'][str(self[id_serveur].id_mod)] = self.stats['serv_modules'][str(self[id_serveur].id_mod)] - 1
            self.stats['serv_variantes'][str(self[id_serveur].id_var)] = self.stats['serv_variantes'][str(self[id_serveur].id_var)] - 1
            del(self[id_serveur])
            if int(id_serveur) in self.stats['no_contact']:
                self.stats['no_contact'].remove(int(id_serveur))
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)

    def edit_serveur(self, id_serveur, dico_modifs):
        # modification d'un serveur dans la base
        serv = self[id_serveur]
        return serv.edit_serveur(dico_modifs)

    def update_contact(self, id_serveur):
        """mise à jour de la date de contact d'un serveur"""
        id_serveur = int(id_serveur)
        cu = cx_pool.create()
        try:
            # on met à jour le serveur dans la base
            query = """update serveurs set last_contact=%s where id=%s"""
            params = (str(time.time()), int(id_serveur))
            cu.execute(query, params)
            cx_pool.commit(cu)
            # mise à jour des statistiques
            if int(id_serveur) in self.stats['no_contact']:
                self.stats['no_contact'].remove(int(id_serveur))
        except:
            traceback.print_exc()
            cx_pool.rollback(cu)


def serveur_pool():
    """crée un pool de tous les serveurs existants
    """
    # création du pool de serveurs en mémoire
    pool=ServeurPool()
    cu = cx_pool.create()
    try:
        cu.execute("select id, rne, libelle, module_actuel, variante, timeout, etat, md5s, maj, no_alert, last_contact from serveurs")
        data = cu.fetchall()
        nbserv = len(data)
        no_contact = []
        # Instanciation des serveurs
        pool.stats['serv_modules'] = {}
        pool.stats['serv_variantes'] = {}
        # initialisation des statistiques des modules/variantes
        for var, module in pool._mod_var.items():
            if str(module) not in pool.stats['serv_modules']:
                pool.stats['serv_modules'][str(module)] = 0
            if str(var) not in pool.stats['serv_variantes']:
                pool.stats['serv_variantes'][str(var)] = 0
        for serv in data:
            # liste des serveurs n'ayant jamais contacté zephir
            if serv[10] == None:
                no_contact.append(int(serv[0]))
            # nombre total de serveurs
            pool[serv[0]] = Serveur(pool, serv[0], cu, serv[1:2])
            pool[serv[0]].update_data(cu, serv[2:10])
            # décompte par module et variante
            pool.stats['serv_modules'][str(serv[3])] = pool.stats['serv_modules'].get(str(serv[3]),0) + 1
            pool.stats['serv_variantes'][str(serv[4])] = pool.stats['serv_variantes'].get(str(serv[4]),0) + 1
        pool.stats['no_contact'] = no_contact
        pool.stats['nb_serv'] = len(pool)
        # chargement des groupes
        pool.update_groupes()
        cx_pool.close(cu)
    except:
        traceback.print_exc()
        cx_pool.close(cu)
    return pool

# utility functions

text_characters = "".join(map(chr, range(32, 255)) + list("\n\r\t\b"))
_null_trans = string.maketrans("", "")

def istextfile(filename, blocksize = 1024):
    return istext(open(filename).read(blocksize))

def istext(s):
    if "\0" in s:
        return 0

    if not s:  # Empty files are considered text
        return 1

    # Get the non-text characters (maps a character to itself then
    # use the 'remove' option to get rid of the text characters.)
    t = s.translate(_null_trans, text_characters)

    # If more than 10% non-text characters, then
    # this is considered a binary file
    if len(t)/len(s) > 0.10:
        return 0
    return 1

